import sys
import time
import random
import threading
from .colorain import *
from .engine import *


class Snake(Sprite):
    def __init__(self, head_init, scene):
        if isinstance(head_init, tuple):
            head_init = Coord2D(*head_init)
        self.positions =  [head_init, Coord2D(head_init.x-1, head_init.y), Coord2D(head_init.x-2, head_init.y)]
        super().__init__(scene, self.positions)
    
    def draw(self):
        self.scene.edit_pixel(self.positions[0], Stx(f"<f=lg;b=b>-</>").parsed)
        for pos in self.positions[1:]:
            self.scene.edit_pixel(pos, Stx(f"<f=lr;b=y>{BLOCKS[2]}</>").parsed)

    def grow(self, direction):
        tail = self.positions[-1]
        if isinstance(direction, Coord2D):
            self.positions.append(direction)
        elif direction == 'VER':
            self.positions.append(Coord2D(tail.x, tail.y-1))
        elif direction == 'HOR':
            self.positions.append(Coord2D(tail.x-1, tail.y))
        elif direction in ['w', 'a', 's', 'd']:
            dx = -1 if direction=='d' else (1 if direction=='a' else 0)
            dy = -1 if direction=='s' else (1 if direction=='w' else 0)
            self.positions.append(Coord2D((tail.x + dx)%self.scene.width, (tail.y + dy)%self.scene.height))
        self.draw()

    def move(self, direction):
        if direction == Coord2D(0,0): 
            return self.positions[-1]
        self.erase()

        predecessor_old_pos = self.positions[0]
        self.positions[0] += direction

        for i in range(1,len(self.positions)):
            curr_node_old_pos = self.positions[i]
            self.positions[i] = predecessor_old_pos
            predecessor_old_pos = curr_node_old_pos
            
        self.draw()
        return curr_node_old_pos


def spawn_apple(scene, snake):
    width = scene.width
    height = scene.height
    ok = False
    while not ok:
        ok = True
        apple_x = random.randint(0, width-1)
        apple_y = random.randint(0, height-1)
        apple_pos = Coord2D(apple_x, apple_y)

        for pos in snake.positions:
            if pos == apple_pos:
                ok = False
    scene.paint_pixel(apple_pos, 'lightred')
    
    return apple_pos

def update_scoreboard(prev_score, scoreboard):
    curr_score = prev_score + 1
    scoreboard_text = f"Score: {curr_score}"
    
    for i in range(len(scoreboard_text)):
        scoreboard.scene[0][i] = Stx(f"<f=b;b=gr>{scoreboard_text[i]}</>").parsed
    
    return curr_score, scoreboard


def key_to_coord(key):
    if key == 'w':
        return Coord2D(0,-1)
    elif key == 's':
        return Coord2D(0, 1)
    elif key == 'a':
        return Coord2D(-1,0)
    elif key == 'd':
        return Coord2D(1, 0)
    else:
        return Coord2D(0,0)


def handle_input(dirn):
    key = ''

    while key != 'q':
        key = getChar()
        if key=='q': 
            dirn[0] =  None
            break
        elif key=='r':
            dirn[1] = 'r'
        else:
            dirn[0] = key_to_coord(key)
    


def snake_game(width = 20, height = 20, start_pos = (5,3), fps = 2):
    gamesc = Scene2D(width, height)
    
    scoreboard = Scene2D(width, 1)
    score, scoreboard = update_scoreboard(-1, scoreboard)

    snake = Snake(Coord2D(*start_pos), gamesc)
    dirn = [key_to_coord('d'), '']
    apple_pos = spawn_apple(gamesc, snake)
    
    inp_thread = threading.Thread(target = handle_input, args=(dirn,))
    inp_thread.start()

    while True:
        scoreboard.render(reset=False)
        gamesc.render()

        if dirn[0] is not None:
            tail_last_pos = snake.move(dirn[0])
            if dirn[1]=='r':
                snake.rotate(math.pi/2)
        else:
            break
    
        if snake.positions[0] == apple_pos:
            score, scoreboard = update_scoreboard(score, scoreboard)
            snake.grow(tail_last_pos)
            apple_pos = spawn_apple(gamesc, snake)
        
        move_cursor(-1,-1)
        time.sleep(1/fps)
    
    return gamesc, snake


##################################################
##################################################
# Submodule for turning the console into a canvas #
# / Console Game Engine #
##################################################
##################################################

##################################################
# Imports #
##################################################
import sys
import math
import threading
from .colorain import *

##################################################
# Block Elements #
##################################################
BLOCK_FULL = chr(9608)
BLOCK_THREEQTR = chr(9619)
BLOCK_HALF = chr(9618)
BLOCK_QTR = chr(9617)

BLOCKS = [BLOCK_FULL, BLOCK_THREEQTR, BLOCK_HALF, BLOCK_QTR]

##################################################
# Move the cursor relative to current position #
##################################################
def move_cursor(line, col):
    base = "\033["

    linedir = "A" if line < 0 else "B"
    coldir = "C" if col > 0 else "D"
    
    linecode = f"{base}{abs(line)}{linedir}" if line!=0 else ""
    colcode = f"{base}{abs(col)}{coldir}" if col!=0 else ""
    sys.stdout.write(linecode)
    sys.stdout.write(colcode) 


##################################################
# Get a one-character input w/o echoing #
##################################################
def getChar():
    try:
        # for Windows-based systems
        import msvcrt # If successful, we are on Windows
        return msvcrt.getch()

    except ImportError:
        # for POSIX-based systems (with termios & tty support)
        import tty, termios  # raises ImportError if unsupported

        fd = sys.stdin.fileno()
        oldSettings = termios.tcgetattr(fd)

        try:
            tty.setcbreak(fd)
            answer = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, oldSettings)

        return answer

def run_input_thread(handle_input, args):
    inp_thread = threading.Thread(target = handle_input, args = args)
    inp_thread.start()

##################################################
# 2D coordinate #
##################################################
class Coord2D:
    def __init__(self, x, y):
        self.x = x
        self.y = y
    
    def __repr__(self):
        return f"{(self.x, self.y)}"
    
    def __add__(self, coord):
        return Coord2D(self.x + coord.x, self.y + coord.y)
    
    def __sub__(self, coord):
        return Coord2D(self.x - coord.x, self.y - coord.y)
    
    def __eq__(self, coord):
        return self.x == coord.x and self.y == coord.y
    
    def dot(self, coord):
        return self.x*coord.x + self.y*coord.y

    def rotate(self, angle, reference):
        rotn_mat = [Coord2D(math.cos(angle), -math.sin(angle)), \
            Coord2D(math.sin(angle), math.cos(angle))]
        newpt = Coord2D(rotn_mat[0].dot(self - reference), rotn_mat[1].dot(self - reference))
        self.x = newpt.x + reference.x
        self.y = newpt.y + reference.y

class Vector2D:
    def __init__(self, start, end):
        if isinstance(start, tuple):
            start = Coord2D(*start)
        if isinstance(end, tuple):
            end = Coord2D(*end)
        self.start = start
        self.end = end
        self.x = self.end.x - self.start.x
        self.y = self.end.y - self.start.y

    def __repr__(self):
        return f"{self.x}i + {self.y}j"
    
    def __add__(self, vector):
        sum_start = self.start
        sum_end = self.end + Coord2D(vector.x, vector.y)
        return Vector2D(sum_start, sum_end)

    def dot(self, vector):
        return self.x*vector.x + self.y*vector.y

    def length(self):
        sqr_len = self.dot(self)
        print(f"{chr(int('221a',16))}{sqr_len}")
        return   math.sqrt(sqr_len)

    def rotate(self, angle):
        return Vector2D(self.start.rotate(angle), self.end.rotate(angle))

##################################################
# 2D scene (i.e., 2D canvas) #
##################################################
class Scene2D:
    def __init__(self, width, height, colour = None):
        self.width = width
        self.height = height
        self.scene = [[BLOCKS[0]][:]*2*self.width for i in range(self.height)]
    
    def render(self, reset=True):
        sys.stdout.write('\n'.join([''.join(ln) for ln in self.scene]))
        if reset:
            move_cursor(-self.height+1, -2*self.width)
        else: print()

    def edit_pixel(self, pixel_loc, newval):
        if isinstance(pixel_loc, Coord2D):
            x, y = pixel_loc.x, pixel_loc.y
        elif isinstance(pixel_loc, tuple):
            x, y = pixel_loc[0], pixel_loc[1]
        self.scene[y][2*x] = newval
        self.scene[y][2*x+1] = newval
    
    def reset_pixel(self, pixel_loc):
        if isinstance(pixel_loc, tuple):
            pixel_loc = Coord2D(pixel_loc[0], pixel_loc[1])
        self.edit_pixel(pixel_loc, BLOCKS[0])

    def paint_pixel(self, pixel_loc, colour):
        if isinstance(pixel_loc, tuple):
            pixel_loc = Coord2D(pixel_loc[0], pixel_loc[1])
        self.edit_pixel(pixel_loc, Stx(f"<{fgtokens[colour]}>{BLOCKS[0]}</>").parsed)


class Sprite:
    def __init__(self, scene, positions):
        self.scene = scene
        self.positions = positions
        self.draw()

    def draw(self, color):
        for pos in self.positions:
            self.scene.paint_pixel(pos, color)
    
    def erase(self):
        for pos in self.positions:
            self.scene.reset_pixel(pos)

    def translate(self, x, y):
        self.erase()
        for pos in self.positions:
            pos.x += x
            pos.y += y
        self.draw()
    
    def rotate(self, angle, reference):
        self.erase()
        for pos in self.positions:
            pos.rotate(angle, reference)
            pos.x = math.floor(pos.x)
            pos.y = math.floor(pos.y)
        self.draw()


class VLine(Sprite):
    def draw(self, color='red'):
        for pos in self.positions:
            self.scene.paint_pixel(pos, color)
    def rotate(self, angle):
        super().rotate(angle, self.positions[0])

class Rect(Sprite):
    def __init__(self, scene, topleft, width, height):
        positions = []
        for i in range(width):
            for j in range(height):
                positions.append(topleft + Coord2D(i,j))
        super().__init__(scene, positions)
    
    def rotate(self, angle):
        super().rotate(angle, self.positions[0])